"""Main hook definition module

All implementations of tox hooks are defined here, as well as any single-use helper functions
specifically related to implementing the hooks (to keep the size/readability of the hook functions
themselves manageable).
"""
from itertools import chain
from typing import Optional

import tox
from tox.action import Action as ToxAction
from tox.config import Parser as ToxParser
from tox.venv import VirtualEnv as ToxVirtualEnv

from tox_poetry_installer import __about__
from tox_poetry_installer import constants
from tox_poetry_installer import exceptions
from tox_poetry_installer import installer
from tox_poetry_installer import logger
from tox_poetry_installer import utilities


def _postprocess_install_project_deps(
    testenv_config, value: Optional[str]  # pylint: disable=unused-argument
) -> Optional[bool]:
    """An awful hack to patch on three-state boolean logic to a config parameter

    .. warning: This logic should 100% be removed in the next feature release. It's here to work
                around a bad design for now but should not persist.

    The bug filed in `#61`_ is caused by a combination of poor design and attempted cleverness. The
    name of the ``install_project_deps`` config option implies that it has ultimate control over
    whether the project dependencies are installed to the testenv, but this is not actually correct.
    What it actually allows the user to do is force the project dependencies to not be installed to
    an environment that would otherwise install them. This was intended behavior, however the
    intention was wrong.

    .. _`#61`: https://github.com/enpaul/tox-poetry-installer/issues/61

    In an effort to be clever the plugin automatically skips installing project dependencies when
    the project package is not installed to the testenv (``skip_install = true``) or if packaging
    as a whole is disabled (``skipsdist = true``). The intention of this behavior is to install only
    the expected dependencies to a testenv and no more. However, this conflicts with the
    ``install_project_deps`` config option, which cannot override this behavior because it defaults
    to ``True``. In effect, ``install_project_deps = true`` in fact means "automatically
    determine whether to install project dependencies" and ``install_project_deps = false`` means
    "never install the project dependencies". This is not ideal and unintuitive.

    To avoid having to make a breaking change this workaround has been added to support three-state
    logic between ``True``, ``False``, and ``None``. The ``install_project_deps`` option is now
    parsed by Tox as a string with a default value of ``None``. If the value is not ``None`` then
    this post processing function will try to convert it to a boolean the same way that Tox's
    `SectionReader.getbool()`_ method does, raising an error to mimic the default behavior if it
    can't.

    .. _`SectionReader.getbool()`: https://github.com/tox-dev/tox/blob/f8459218ee5ab5753321b3eb989b7beee5b391ad/src/tox/config/__init__.py#L1724

    The three states for the ``install_project_deps`` setting are:
    * ``None`` - User did not configure the setting, package dependency installation is
      determined automatically
    * ``True`` - User configured the setting to ``True``, package dependencies will be installed
    * ``False`` - User configured the setting to ``False``, package dependencies will not be
      installed

    This config option should be deprecated with the 1.0.0 release and instead an option like
    ``always_install_project_deps`` should be added which overrides the default determination and
    just installs the project dependencies. The counterpart (``never_install_project_deps``)
    shouldn't be needed, since I don't think there's a real use case for that.
    """
    if value is None:
        return value

    if value.lower() == "true":
        return True
    if value.lower() == "false":
        return False

    raise tox.exception.ConfigError(
        f"install_project_deps: boolean value '{value}' needs to be 'True' or 'False'"
    )


@tox.hookimpl
def tox_addoption(parser: ToxParser):
    """Add required configuration options to the tox INI file

    Adds the ``require_locked_deps`` configuration option to the venv to check whether all
    dependencies should be treated as locked or not.
    """

    parser.add_argument(
        "--require-poetry",
        action="store_true",
        dest="require_poetry",
        help="(deprecated) Trigger a failure if Poetry is not available to Tox",
    )

    parser.add_argument(
        "--parallelize-locked-install",
        type=int,
        dest="parallelize_locked_install",
        default=None,
        help="(deprecated) Number of worker threads to use for installing dependencies from the Poetry lockfile in parallel",
    )

    parser.add_argument(
        "--parallel-install-threads",
        type=int,
        dest="parallel_install_threads",
        default=constants.DEFAULT_INSTALL_THREADS,
        help="Number of locked dependencies to install simultaneously; set to 0 to disable parallel installation",
    )

    parser.add_testenv_attribute(
        name="install_dev_deps",
        type="bool",
        default=False,
        help="(deprecated) Automatically install all Poetry development dependencies to the environment",
    )

    parser.add_testenv_attribute(
        name="poetry_dep_groups",
        type="line-list",
        default=[],
        help="List of Poetry dependency groups to install to the environment",
    )

    parser.add_testenv_attribute(
        name="install_project_deps",
        type="string",
        default=None,
        help="Automatically install all Poetry primary dependencies to the environment",
        postprocess=_postprocess_install_project_deps,
    )

    parser.add_testenv_attribute(
        name="require_locked_deps",
        type="bool",
        default=False,
        help="Require all dependencies in the environment be installed using the Poetry lockfile",
    )

    parser.add_testenv_attribute(
        name="require_poetry",
        type="bool",
        default=False,
        help="Trigger a failure if Poetry is not available to Tox",
    )

    parser.add_testenv_attribute(
        name="locked_deps",
        type="line-list",
        help="List of locked dependencies to install to the environment using the Poetry lockfile",
    )


@tox.hookimpl
def tox_testenv_install_deps(venv: ToxVirtualEnv, action: ToxAction) -> Optional[bool]:
    """Install the dependencies for the current environment

    Loads the local Poetry environment and the corresponding lockfile then pulls the dependencies
    specified by the Tox environment. Finally these dependencies are installed into the Tox
    environment using the Poetry ``PipInstaller`` backend.

    :param venv: Tox virtual environment object with configuration for the local Tox environment.
    :param action: Tox action object
    """

    try:
        poetry = utilities.check_preconditions(venv, action)
    except exceptions.SkipEnvironment as err:
        if isinstance(err, exceptions.PoetryNotInstalledError) and (
            venv.envconfig.config.option.require_poetry or venv.envconfig.require_poetry
        ):
            venv.status = err.__class__.__name__
            logger.error(str(err))
            return False
        logger.info(str(err))
        return None

    logger.info(f"Loaded project pyproject.toml from {poetry.file}")

    virtualenv = utilities.convert_virtualenv(venv)

    if not poetry.locker.is_fresh():
        logger.warning(
            f"The Poetry lock file is not up to date with the latest changes in {poetry.file}"
        )

    try:
        if venv.envconfig.require_locked_deps and venv.envconfig.deps:
            raise exceptions.LockedDepsRequiredError(
                f"Unlocked dependencies '{venv.envconfig.deps}' specified for environment '{venv.name}' which requires locked dependencies"
            )

        packages = utilities.build_package_map(poetry)

        if venv.envconfig.install_dev_deps:
            dev_deps = utilities.find_dev_deps(packages, virtualenv, poetry)
            logger.info(
                f"Identified {len(dev_deps)} development dependencies to install to env"
            )
        else:
            dev_deps = []
            logger.info("Env does not install development dependencies, skipping")

        group_deps = utilities.dedupe_packages(
            list(
                chain(
                    *[
                        utilities.find_group_deps(group, packages, virtualenv, poetry)
                        for group in venv.envconfig.poetry_dep_groups
                    ]
                )
            )
        )
        logger.info(
            f"Identified {len(group_deps)} group dependencies to install to env"
        )

        env_deps = utilities.find_additional_deps(
            packages, virtualenv, poetry, venv.envconfig.locked_deps
        )

        logger.info(
            f"Identified {len(env_deps)} environment dependencies to install to env"
        )

        install_project_deps = (
            venv.envconfig.install_project_deps
            if venv.envconfig.install_project_deps is not None
            else (
                not venv.envconfig.skip_install and not venv.envconfig.config.skipsdist
            )
        )

        if install_project_deps:
            project_deps = utilities.find_project_deps(
                packages, virtualenv, poetry, venv.envconfig.extras
            )
            logger.info(
                f"Identified {len(project_deps)} project dependencies to install to env"
            )
        else:
            project_deps = []
            logger.info("Env does not install project package dependencies, skipping")
    except exceptions.ToxPoetryInstallerException as err:
        venv.status = err.__class__.__name__
        logger.error(str(err))
        return False
    except Exception as err:
        venv.status = "InternalError"
        logger.error(f"Internal plugin error: {err}")
        raise err

    dependencies = utilities.dedupe_packages(
        dev_deps + group_deps + env_deps + project_deps
    )
    if (
        venv.envconfig.config.option.parallel_install_threads
        != constants.DEFAULT_INSTALL_THREADS
    ):
        parallel_threads = venv.envconfig.config.option.parallel_install_threads
    else:
        parallel_threads = (
            venv.envconfig.config.option.parallelize_locked_install
            if venv.envconfig.config.option.parallelize_locked_install is not None
            else constants.DEFAULT_INSTALL_THREADS
        )
    log_parallel = f" (using {parallel_threads} threads)" if parallel_threads else ""

    action.setactivity(
        __about__.__title__,
        f"Installing {len(dependencies)} dependencies from Poetry lock file{log_parallel}",
    )
    installer.install(
        poetry,
        venv,
        dependencies,
        parallel_threads,
    )

    return venv.envconfig.require_locked_deps or None



# pylint: disable=missing-module-docstring, missing-function-docstring, unused-argument, too-few-public-methods
import time
from pathlib import Path

import poetry.factory
import poetry.installation.pip_installer
import poetry.utils.env
import pytest
import tox
from poetry.core.packages.package import Package as PoetryPackage

from tox_poetry_installer import utilities


TEST_PROJECT_PATH = Path(__file__).parent.resolve() / "test-project"

FAKE_VENV_PATH = Path("nowhere")


class MockVirtualEnv:
    """Mock class for the :class:`poetry.utils.env.VirtualEnv` and :class:`tox.venv.VirtualEnv`"""

    class MockTestenvConfig:  # pylint: disable=missing-class-docstring
        envdir = FAKE_VENV_PATH

    def __init__(self, *args, **kwargs):
        self.envconfig = self.MockTestenvConfig()
        self.installed = []

    @staticmethod
    def is_valid_for_marker(*args, **kwargs):
        return True

    @staticmethod
    def get_version_info():
        return (1, 2, 3)


class MockPipInstaller:
    """Mock class for the :class:`poetry.installation.pip_installer.PipInstaller`"""

    def __init__(self, env: MockVirtualEnv, **kwargs):
        self.env = env

    def install(self, package: PoetryPackage):
        self.env.installed.append(package)
        time.sleep(1)


@pytest.fixture
def mock_venv(monkeypatch):
    monkeypatch.setattr(utilities, "convert_virtualenv", lambda venv: venv)
    monkeypatch.setattr(
        poetry.installation.pip_installer, "PipInstaller", MockPipInstaller
    )
    monkeypatch.setattr(tox.venv, "VirtualEnv", MockVirtualEnv)
    monkeypatch.setattr(poetry.utils.env, "VirtualEnv", MockVirtualEnv)


@pytest.fixture(scope="function")
def mock_poetry_factory(monkeypatch):
    pypoetry = poetry.factory.Factory().create_poetry(cwd=TEST_PROJECT_PATH)

    def mock_factory(*args, **kwargs):
        return pypoetry

    monkeypatch.setattr(poetry.factory.Factory, "create_poetry", mock_factory)



"""Helper utility functions, usually bridging Tox and Poetry functionality"""
# Silence this one globally to support the internal function imports for the proxied poetry module.
# See the docstring in 'tox_poetry_installer._poetry' for more context.
# pylint: disable=import-outside-toplevel
import collections
import typing
from pathlib import Path
from typing import Dict
from typing import List
from typing import Sequence
from typing import Set

from poetry.core.packages.dependency import Dependency as PoetryDependency
from poetry.core.packages.package import Package as PoetryPackage
from tox.action import Action as ToxAction
from tox.venv import VirtualEnv as ToxVirtualEnv

from tox_poetry_installer import constants
from tox_poetry_installer import exceptions
from tox_poetry_installer import logger

if typing.TYPE_CHECKING:
    from tox_poetry_installer import _poetry


PackageMap = Dict[str, List[PoetryPackage]]


def check_preconditions(venv: ToxVirtualEnv, action: ToxAction) -> "_poetry.Poetry":
    """Check that the local project environment meets expectations"""
    # Skip running the plugin for the provisioning environment. The provisioned environment,
    # for alternative Tox versions and/or the ``requires`` meta dependencies is specially
    # handled by Tox and is out of scope for this plugin. Since one of the ways to install this
    # plugin in the first place is via the Tox provisioning environment, it quickly becomes a
    # chicken-and-egg problem.
    if action.name == venv.envconfig.config.provision_tox_env:
        raise exceptions.SkipEnvironment(
            f"Skipping Tox provisioning env '{action.name}'"
        )

    # Skip running the plugin for the packaging environment. PEP-517 front ends can handle
    # that better than we can, so let them do their thing. More to the point: if you're having
    # problems in the packaging env that this plugin would solve, god help you.
    if action.name == venv.envconfig.config.isolated_build_env:
        raise exceptions.SkipEnvironment(
            f"Skipping isolated packaging build env '{action.name}'"
        )

    if venv.envconfig.config.option.require_poetry:
        logger.warning(
            "DEPRECATION: The '--require-poetry' runtime option is deprecated and will be "
            "removed in version 1.0.0. Please update test environments that require Poetry to "
            "set the 'require_poetry = true' option in tox.ini"
        )

    if venv.envconfig.config.option.parallelize_locked_install is not None:
        logger.warning(
            "DEPRECATION: The '--parallelize-locked-install' option is deprecated and will "
            "be removed in version 1.0.0. Please use the '--parallel-install-threads' option."
        )

    if venv.envconfig.install_dev_deps:
        logger.warning(
            "DEPRECATION: The 'install_dev_deps' option is deprecated and will be removed in "
            "version 1.0.0. Please update test environments that install development dependencies "
            "to set the 'poetry_dev_groups = [dev]' option in tox.ini"
        )

    from tox_poetry_installer import _poetry

    try:
        return _poetry.Factory().create_poetry(venv.envconfig.config.toxinidir)
    # Support running the plugin when the current tox project does not use Poetry for its
    # environment/dependency management.
    #
    # ``RuntimeError`` is dangerous to blindly catch because it can be (and in Poetry's case,
    # is) raised in many different places for different purposes.
    except RuntimeError:
        raise exceptions.SkipEnvironment(
            "Project does not use Poetry for env management, skipping installation of locked dependencies"
        ) from None


def convert_virtualenv(venv: ToxVirtualEnv) -> "_poetry.VirtualEnv":
    """Convert a Tox venv to a Poetry venv

    :param venv: Tox ``VirtualEnv`` object representing a tox virtual environment
    :returns: Poetry ``VirtualEnv`` object representing a poetry virtual environment
    """
    from tox_poetry_installer import _poetry

    return _poetry.VirtualEnv(path=Path(venv.envconfig.envdir))


def build_package_map(poetry: "_poetry.Poetry") -> PackageMap:
    """Build the mapping of package names to objects

    :param poetry: Populated poetry object to load locked packages from
    :returns: Mapping of package names to Poetry package objects
    """
    packages = collections.defaultdict(list)
    for package in poetry.locker.locked_repository().packages:
        packages[package.name].append(package)

    return packages


def identify_transients(
    dep_name: str,
    packages: PackageMap,
    venv: "_poetry.VirtualEnv",
    allow_missing: Sequence[str] = (),
) -> List[PoetryPackage]:
    """Using a pool of packages, identify all transient dependencies of a given package name

    :param dep_name: Either the Poetry dependency or the dependency's bare package name to recursively
                     identify the transient dependencies of
    :param packages: All packages from the lockfile to use for identifying dependency relationships.
    :param venv: Poetry virtual environment to use for package compatibility checks
    :param allow_missing: Sequence of package names to allow to be missing from the lockfile. Any
                          packages that are not found in the lockfile but their name appears in this
                          list will be silently skipped from installation.
    :returns: List of packages that need to be installed for the requested dependency.

    .. note:: The package corresponding to the dependency specified by the ``dep`` parameter will
              be included in the returned list of packages.
    """
    searched: Set[str] = set()

    def _transients(transient: PoetryDependency) -> List[PoetryPackage]:
        searched.add(transient.name)

        results: List[PoetryPackage] = []
        for option in packages[transient.name]:
            if venv.is_valid_for_marker(option.to_dependency().marker):
                for requirement in option.requires:
                    if requirement.name not in searched:
                        results += _transients(requirement)
                logger.debug(f"Including {option} for installation")
                results.append(option)
                break
        else:
            logger.debug(
                f"Skipping {transient.name}: target python version is {'.'.join([str(item) for item in venv.get_version_info()])} but package requires {transient.marker}"
            )

        return results

    try:
        for option in packages[dep_name]:
            if venv.is_valid_for_marker(option.to_dependency().marker):
                dep = option.to_dependency()
                break
        else:
            logger.warning(
                f"Skipping {dep_name}: no locked version found compatible with target python version {'.'.join([str(item) for item in venv.get_version_info()])}"
            )
            return []

        return _transients(dep)
    except KeyError as err:
        missing = err.args[0]

        if missing in constants.UNSAFE_PACKAGES:
            logger.warning(
                f"Installing package '{missing}' using Poetry is not supported and will be skipped"
            )
            logger.debug(f"Skipping {missing}: designated unsafe by Poetry")
            return []

        if missing in allow_missing:
            logger.debug(f"Skipping {missing}: package is allowed to be unlocked")
            return []

        if any(
            delimiter in missing for delimiter in constants.PEP508_VERSION_DELIMITERS
        ):
            raise exceptions.LockedDepVersionConflictError(
                f"Locked dependency '{missing}' cannot include version specifier"
            ) from None

        raise exceptions.LockedDepNotFoundError(
            f"No version of locked dependency '{missing}' found in the project lockfile"
        ) from None


def find_project_deps(
    packages: PackageMap,
    venv: "_poetry.VirtualEnv",
    poetry: "_poetry.Poetry",
    extras: Sequence[str] = (),
) -> List[PoetryPackage]:
    """Find the root project dependencies

    Recursively identify the dependencies of the root project package

    :param packages: Mapping of all locked package names to their corresponding package object
    :param venv: Poetry virtual environment to use for package compatibility checks
    :param poetry: Poetry object for the current project
    :param extras: Sequence of extra names to include the dependencies of
    """

    if any(dep.name in constants.UNSAFE_PACKAGES for dep in poetry.package.requires):
        raise exceptions.RequiresUnsafeDepError(
            f"Project package requires one or more unsafe dependencies ({', '.join(constants.UNSAFE_PACKAGES)}) which cannot be installed with Poetry"
        )

    required_dep_names = [
        item.name for item in poetry.package.requires if not item.is_optional()
    ]

    extra_dep_names: List[str] = []
    for extra in extras:
        logger.info(f"Processing project extra '{extra}'")
        try:
            extra_dep_names += [item.name for item in poetry.package.extras[extra]]
        except KeyError:
            raise exceptions.ExtraNotFoundError(
                f"Environment specifies project extra '{extra}' which was not found in the lockfile"
            ) from None

    dependencies: List[PoetryPackage] = []
    for dep_name in required_dep_names + extra_dep_names:
        dependencies += identify_transients(
            dep_name.lower(), packages, venv, allow_missing=[poetry.package.name]
        )

    return dedupe_packages(dependencies)


def find_additional_deps(
    packages: PackageMap,
    venv: "_poetry.VirtualEnv",
    poetry: "_poetry.Poetry",
    dep_names: Sequence[str],
) -> List[PoetryPackage]:
    """Find additional dependencies

    Recursively identify the dependencies of an arbitrary list of package names

    :param packages: Mapping of all locked package names to their corresponding package object
    :param venv: Poetry virtual environment to use for package compatibility checks
    :param poetry: Poetry object for the current project
    :param dep_names: Sequence of additional dependency names to recursively find the transient
                      dependencies for
    """
    dependencies: List[PoetryPackage] = []
    for dep_name in dep_names:
        dependencies += identify_transients(
            dep_name.lower(), packages, venv, allow_missing=[poetry.package.name]
        )

    return dedupe_packages(dependencies)


def find_group_deps(
    group: str,
    packages: PackageMap,
    venv: "_poetry.VirtualEnv",
    poetry: "_poetry.Poetry",
) -> List[PoetryPackage]:
    """Find the dependencies belonging to a dependency group

    Recursively identify the Poetry dev dependencies

    :param group: Name of the dependency group from the project's ``pyproject.toml``
    :param packages: Mapping of all locked package names to their corresponding package object
    :param venv: Poetry virtual environment to use for package compatibility checks
    :param poetry: Poetry object for the current project
    """
    return find_additional_deps(
        packages,
        venv,
        poetry,
        poetry.pyproject.data["tool"]["poetry"]
        .get("group", {})
        .get(group, {})
        .get("dependencies", {})
        .keys(),
    )


def find_dev_deps(
    packages: PackageMap, venv: "_poetry.VirtualEnv", poetry: "_poetry.Poetry"
) -> List[PoetryPackage]:
    """Find the dev dependencies

    Recursively identify the Poetry dev dependencies

    :param packages: Mapping of all locked package names to their corresponding package object
    :param venv: Poetry virtual environment to use for package compatibility checks
    :param poetry: Poetry object for the current project
    """
    dev_group_deps = find_group_deps("dev", packages, venv, poetry)

    # Legacy pyproject.toml poetry format:
    legacy_dev_group_deps = find_additional_deps(
        packages,
        venv,
        poetry,
        poetry.pyproject.data["tool"]["poetry"].get("dev-dependencies", {}).keys(),
    )

    # Poetry 1.2 unions these two toml sections.
    return dedupe_packages(dev_group_deps + legacy_dev_group_deps)


def dedupe_packages(packages: Sequence[PoetryPackage]) -> List[PoetryPackage]:
    """Deduplicates a sequence of PoetryPackages while preserving ordering

    Adapted from StackOverflow: https://stackoverflow.com/a/480227
    """
    seen: Set[PoetryPackage] = set()
    # Make this faster, avoid method lookup below
    seen_add = seen.add
    return [p for p in packages if not (p in seen or seen_add(p))]



# pylint: disable=missing-module-docstring, redefined-outer-name, unused-argument, wrong-import-order, unused-import
import time

import tox.venv
from poetry.factory import Factory

from .fixtures import mock_poetry_factory
from .fixtures import mock_venv
from tox_poetry_installer import installer
from tox_poetry_installer import utilities


def test_deduplication(mock_venv, mock_poetry_factory):
    """Test that the installer does not install duplicate dependencies"""
    poetry = Factory().create_poetry(None)
    packages: utilities.PackageMap = {
        item.name: item for item in poetry.locker.locked_repository().packages
    }

    venv = tox.venv.VirtualEnv()
    to_install = [packages["toml"], packages["toml"]]

    installer.install(poetry, venv, to_install)

    assert len(set(to_install)) == len(venv.installed)  # pylint: disable=no-member


def test_parallelization(mock_venv, mock_poetry_factory):
    """Test that behavior is consistent between parallel and non-parallel usage"""
    poetry = Factory().create_poetry(None)
    packages: utilities.PackageMap = {
        item.name: item for item in poetry.locker.locked_repository().packages
    }

    to_install = [
        packages["toml"],
        packages["toml"],
        packages["tox"],
        packages["requests"],
        packages["python-dateutil"],
        packages["attrs"],
    ]

    venv_sequential = tox.venv.VirtualEnv()
    start_sequential = time.time()
    installer.install(poetry, venv_sequential, to_install, 0)
    sequential = time.time() - start_sequential

    venv_parallel = tox.venv.VirtualEnv()
    start_parallel = time.time()
    installer.install(poetry, venv_parallel, to_install, 5)
    parallel = time.time() - start_parallel

    # The mock delay during package install is static (one second) so these values should all
    # be within microseconds of each other
    assert parallel < sequential
    assert round(parallel * 5) == round(sequential)
    assert round(sequential) == len(set(to_install))
    assert round(parallel * 5) == len(set(to_install))



# pylint: disable=missing-module-docstring, redefined-outer-name, unused-argument, wrong-import-order, unused-import
import poetry.factory
import poetry.utils.env
import pytest
from poetry.puzzle.provider import Provider

from .fixtures import mock_poetry_factory
from .fixtures import mock_venv
from tox_poetry_installer import constants
from tox_poetry_installer import exceptions
from tox_poetry_installer import utilities


def test_exclude_unsafe():
    """Test that the unsafe packages are properly excluded

    Also ensure that the internal constant matches the value from Poetry
    """
    assert Provider.UNSAFE_PACKAGES == constants.UNSAFE_PACKAGES

    for dep in constants.UNSAFE_PACKAGES:
        assert not utilities.identify_transients(dep, {}, None)


def test_allow_missing():
    """Test that the ``allow_missing`` parameter works as expected"""
    with pytest.raises(exceptions.LockedDepNotFoundError):
        utilities.identify_transients("luke-skywalker", {}, None)

    assert not utilities.identify_transients(
        "darth-vader", {}, None, allow_missing=["darth-vader"]
    )


def test_exclude_pep508():
    """Test that dependencies specified in PEP508 format are properly excluded"""
    for version in [
        "foo==1.0",
        "foo==1",
        "foo>2.0.0",
        "foo<=9.3.4.7.8",
        "foo>999,<=4.6",
        "foo>1234 || foo<2021.01.01",
        "foo!=7",
        "foo~=0.8",
        "foo!=9,==7",
        "=>foo",
    ]:
        with pytest.raises(exceptions.LockedDepVersionConflictError):
            utilities.identify_transients(version, {}, None)


def test_functional(mock_poetry_factory, mock_venv):
    """Integration tests for the :func:`identify_transients` function

    Trivially test that it resolves dependencies properly and that the parent package
    is always the last in the returned list.
    """
    pypoetry = poetry.factory.Factory().create_poetry(None)
    packages = utilities.build_package_map(pypoetry)
    venv = poetry.utils.env.VirtualEnv()  # pylint: disable=no-value-for-parameter

    requests_requires = [
        packages["certifi"][0],
        packages["chardet"][0],
        packages["idna"][0],
        packages["urllib3"][0],
        packages["requests"][0],
    ]

    transients = utilities.identify_transients("requests", packages, venv)

    assert all((item in requests_requires) for item in transients)
    assert all((item in transients) for item in requests_requires)

    for package in [packages["requests"][0], packages["tox"][0], packages["flask"][0]]:
        transients = utilities.identify_transients(package.name, packages, venv)
        assert transients[-1] == package
        assert len(transients) == len(set(transients))



"""Ensure that the pyproject and module metadata never drift out of sync

The next best thing to having one source of truth is having a way to ensure all of your
sources of truth agree with each other.
"""
from pathlib import Path

import toml

from tox_poetry_installer import __about__


def test_metadata():
    """Test that module metadata matches pyproject poetry metadata"""

    with (Path(__file__).resolve().parent / ".." / "pyproject.toml").open() as infile:
        pyproject = toml.load(infile, _dict=dict)

    assert pyproject["tool"]["poetry"]["name"] == __about__.__title__
    assert pyproject["tool"]["poetry"]["version"] == __about__.__version__
    assert pyproject["tool"]["poetry"]["license"] == __about__.__license__
    assert pyproject["tool"]["poetry"]["description"] == __about__.__summary__
    assert pyproject["tool"]["poetry"]["repository"] == __about__.__url__
    assert (
        all(
            item in __about__.__authors__
            for item in pyproject["tool"]["poetry"]["authors"]
        )
        is True
    )
    assert (
        all(
            item in pyproject["tool"]["poetry"]["authors"]
            for item in __about__.__authors__
        )
        is True
    )



"""Funcationality for performing virtualenv installation"""
# Silence this one globally to support the internal function imports for the proxied poetry module.
# See the docstring in 'tox_poetry_installer._poetry' for more context.
# pylint: disable=import-outside-toplevel
import concurrent.futures
import contextlib
import typing
from datetime import datetime
from typing import Collection
from typing import Set

from poetry.core.packages.package import Package as PoetryPackage
from tox.venv import VirtualEnv as ToxVirtualEnv

from tox_poetry_installer import logger
from tox_poetry_installer import utilities

if typing.TYPE_CHECKING:
    from tox_poetry_installer import _poetry


def install(
    poetry: "_poetry.Poetry",
    venv: ToxVirtualEnv,
    packages: Collection[PoetryPackage],
    parallels: int = 0,
):
    """Install a bunch of packages to a virtualenv

    :param poetry: Poetry object the packages were sourced from
    :param venv: Tox virtual environment to install the packages to
    :param packages: List of packages to install to the virtual environment
    :param parallels: Number of parallel processes to use for installing dependency packages, or
                      ``None`` to disable parallelization.
    """
    from tox_poetry_installer import _poetry

    logger.info(
        f"Installing {len(packages)} packages to environment at {venv.envconfig.envdir}"
    )

    pip = _poetry.PipInstaller(
        env=utilities.convert_virtualenv(venv),
        io=_poetry.NullIO(),
        pool=poetry.pool,
    )

    installed: Set[PoetryPackage] = set()

    def logged_install(dependency: PoetryPackage) -> None:
        start = datetime.now()
        logger.debug(f"Installing {dependency}")
        pip.install(dependency)
        end = datetime.now()
        logger.debug(f"Finished installing {dependency} in {end - start}")

    @contextlib.contextmanager
    def _optional_parallelize():
        """A bit of cheat, really

        A context manager that exposes a common interface for the caller that optionally
        enables/disables the usage of the parallel thread pooler depending on the value of
        the ``parallels`` parameter.
        """
        if parallels > 0:
            with concurrent.futures.ThreadPoolExecutor(
                max_workers=parallels
            ) as executor:
                yield executor.submit
        else:
            yield lambda func, arg: func(arg)

    with _optional_parallelize() as executor:
        for dependency in packages:
            if dependency not in installed:
                installed.add(dependency)
                logger.debug(f"Queuing {dependency}")
                executor(logged_install, dependency)
            else:
                logger.debug(f"Skipping {dependency}, already installed")
        logger.debug("Waiting for installs to finish...")



"""Logging wrappers to reduce duplication elsewhere

Calling ``tox.reporter.something()`` and having to format a string with the prefix
gets really old fast, but more importantly it also makes the flow of the main code
more difficult to follow because of the added complexity.
"""
import tox

from tox_poetry_installer import constants


def error(message: str):
    """Wrapper around :func:`tox.reporter.error`"""
    tox.reporter.error(f"{constants.REPORTER_PREFIX} {message}")


def warning(message: str):
    """Wrapper around :func:`tox.reporter.warning`"""
    tox.reporter.warning(f"{constants.REPORTER_PREFIX} {message}")


def info(message: str):
    """Wrapper around :func:`tox.reporter.verbosity1`"""
    tox.reporter.verbosity1(f"{constants.REPORTER_PREFIX} {message}")


def debug(message: str):
    """Wrapper around :func:`tox.reporter.verbosity2`"""
    tox.reporter.verbosity2(f"{constants.REPORTER_PREFIX} {message}")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tests', 'tests.test-project.test_project', 'tox_poetry_installer']

package_data = \
{'': ['*'], 'tests': ['test-project/*']}

install_requires = \
['poetry-core>=1.1.0,<2.0.0', 'tox>=3.8.0,<4.0.0']

extras_require = \
{'poetry': ['cleo>=1.0,<3.0', 'poetry>=1.2.0,<2.0.0']}

entry_points = \
{'tox': ['poetry_installer = tox_poetry_installer']}

setup_kwargs = {
    'name': 'tox-poetry-installer',
    'version': '0.10.1',
    'description': 'A plugin for Tox that lets you install test environment dependencies from the Poetry lockfile',
    'long_description': '# tox-poetry-installer\n\nA plugin for [Tox](https://tox.readthedocs.io/en/latest/) that lets you install test\nenvironment dependencies from the [Poetry](https://python-poetry.org/) lockfile.\n\n[![CI Status](https://github.com/enpaul/tox-poetry-installer/workflows/CI/badge.svg?event=push)](https://github.com/enpaul/tox-poetry-installer/actions)\n[![PyPI Version](https://img.shields.io/pypi/v/tox-poetry-installer)](https://pypi.org/project/tox-poetry-installer/)\n[![PyPI Downloads](https://img.shields.io/pypi/dm/tox-poetry-installer)](https://libraries.io/pypi/tox-poetry-installer)\n[![License](https://img.shields.io/pypi/l/tox-poetry-installer)](https://opensource.org/licenses/MIT)\n[![Python Supported Versions](https://img.shields.io/pypi/pyversions/tox-poetry-installer)](https://www.python.org)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n\n⚠️ **This project is beta software and is under active development** ⚠️\n\n## Documentation\n\n- [Feature Overview](#feature-overview)\n- [Using the Plugin](#user-documentation)\n  - [Installing](#installing)\n  - [Quick Start](#quick-start)\n  - [References](#references)\n    - [Config Options](#configuration-options)\n    - [Runtime Options](#runtime-options)\n    - [Errors](#errors)\n  - [Other Notes](#other-notes)\n    - [Unsupported Tox config options](#unsupported-tox-config-options)\n    - [Updating locked dependencies in a testenv](#updating-locked-dependencies-in-a-testenv)\n    - [Installing unsafe dependencies](#installing-unsafe-dependencies)\n    - [Using with an unmanaged Poetry installation](#using-with-an-unmanaged-poetry-installation)\n- [Developing the Plugin](#developer-documentation)\n- [Road Map](#road-map)\n\nSee the\n[Changelog](https://github.com/enpaul/tox-poetry-installer/blob/devel/CHANGELOG.md) for\nrelease history.\n\n*See also: [official Tox plugins](https://tox.readthedocs.io/en/latest/plugins.html), [Poetry Tox plugin](https://github.com/tkukushkin/tox-poetry)*\n\n## Feature Overview\n\n- Manage package versions in exactly one place and with exactly one tool: Poetry.\n- Ensure CI/CD and other automation tools are using the same package versions that you are\n  in your local development environment.\n- Add only the packages or custom groups you need to a Tox test environment, instead of\n  everything in your lockfile.\n- Directly integrate with Poetry, re-using your existing package indexes and credentials,\n  with no additional configuration.\n- Wherever possible, built-in Tox config options are always respected and their behavior\n  kept consistent.\n- Extremely configurable. Every feature can be disabled or enabled for any given Tox test\n  environment.\n- Friendly to other Tox plugins and supports a wide range of environments.\n\n## User Documentation\n\n*This section is for users looking to integrate the plugin with their project or CI system. For information on contributing to the plugin please see the [Developer Docs](#developer-documentation)*\n\n### Installing\n\nThe recommended way to install the plugin is to add it to a project using Poetry:\n\n```bash\npoetry add tox-poetry-installer[poetry] --dev\n```\n\n> **Note:** Always install the plugin with the `[poetry]` extra, unless you are\n> [managing the Poetry installation yourself](#externally-managed-poetry-installation).\n\nAlternatively, it can be installed directly to a virtual environment using Pip, though\nthis is not recommended:\n\n```bash\nsource somevenv/bin/activate\npip install tox-poetry-installer\n```\n\nAlternatively alternatively, it can be installed using the Tox\n[`requires`](https://tox.readthedocs.io/en/latest/config.html#conf-requires) option by\nadding the below to `tox.ini`, though this is also not recommended:\n\n```ini\nrequires =\n    tox-poetry-installer[poetry] == 0.8.0\n```\n\nAfter installing, check that Tox recognizes the plugin by running\n`poetry run tox --version`. The command should give output similar to below:\n\n```\n3.20.0 imported from .venv/lib64/python3.8/site-packages/tox/__init__.py\nregistered plugins:\n    tox-poetry-installer-0.8.0 at .venv/lib64/python3.8/site-packages/tox_poetry_installer/__init__.py\n```\n\n### Quick Start\n\nCongratulations! 🎉 Just by installing the plugin your Tox config is already using locked\ndependencies: when Tox builds and installs your project package to a test environment,\nyour project package\'s dependencies will be installed from the lockfile.\n\nNow lets update an example `tox.ini` to install the other test environment dependencies\nfrom the lockfile.\n\nA `testenv` from the example `tox.ini` we\'re starting with is below:\n\n```ini\n[testenv]\ndescription = Some very cool tests\ndeps =\n    black == 20.8b1\n    pylint >=2.4.4,<2.7.0\n    mypy <0.800\ncommands = ...\n```\n\nTo update the config so that the testenv dependencies are installed from the lockfile, we\ncan replace the built-in\n[`deps`](https://tox.readthedocs.io/en/latest/config.html#conf-deps) option with the\n`locked_deps` option provided by the plugin, and then remove the inline version\nspecifiers. With these changes the three testenv dependencies (as well as all of their\ndependencies) will be installed from the lockfile when the test environment is recreated:\n\n```ini\n[testenv]\ndescription = Some very cool tests\nlocked_deps =\n    black\n    pylint\n    mypy\ncommands = ...\n```\n\nWe can also add the `require_locked_deps` option to the test environment. This will both\nblock any other install tools (another plugin or Tox itself) from installing dependencies\nto the Tox environment and also cause Tox to fail if the test environment also uses the\nbuilt-in [`deps`](https://tox.readthedocs.io/en/latest/config.html#conf-deps) option:\n\n```ini\n[testenv]\ndescription = Some very cool tests\nrequire_locked_deps = true\nlocked_deps =\n    black\n    pylint\n    mypy\ncommands = ...\n```\n\n> **Note:** Settings configured on the main `testenv` environment are inherited by child\n> test environments (for example, `testenv:foo`). To override this, specify the setting in\n> the child environment with a different value.\n\nAlternatively, we can skip specifying all of our dependencies for a test environment in\nthe Tox config and install Poetry dependency groups directly:\n\n```ini\n[testenv]\ndescription = Some very cool tests\nrequire_locked_deps = true\npoetry_dep_groups =\n    dev\ncommands = ...\n```\n\n> **Note:** Setting `poetry_dep_groups = [dev]` on an environment that also installs the\n> project package is functionally equivalent to running `poetry install`.\n\n> **Note:** The `install_dev_deps` configuration option is deprecated. See [Configuration\n> Options](#configuration-options) for more information.\n\nFinally, we can also install an unlocked dependency (a dependency which doesn\'t take its\nversion from the Poetry lockfile) into the test environment alongside the locked ones. We\nneed to remove the `require_locked_deps = true` option, otherwise the environment will\nerror, and then we can add the unlocked dependency using the built-in\n[`deps`](https://tox.readthedocs.io/en/latest/config.html#conf-deps) option:\n\n```ini\n[testenv]\ndescription = Some very cool tests\ndeps =\n    pytest >= 5.6.0,<6.0.0\nlocked_deps =\n    black\n    pylint\n    mypy\ncommands = ...\n```\n\n## References\n\n### Configuration Options\n\nAll options listed below are Tox environment options and can be applied to one or more\nenvironment sections of the `tox.ini` file. They cannot be applied to the global Tox\nconfiguration section.\n\n> **Note:** Settings configured on the main `testenv` environment are inherited by child\n> test environments (for example, `testenv:foo`). To override this, specify the setting in\n> the child environment with a different value.\n\n| Option                 |  Type   | Default | Description                                                                                                                                                                                                                                                                                                                                                          |\n| :--------------------- | :-----: | :-----: | :------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------- |\n| `locked_deps`          |  List   |  `[]`   | Names of packages to install to the test environment from the Poetry lockfile. Transient dependencies (packages required by these dependencies) are automatically included.                                                                                                                                                                                          |\n| `require_locked_deps`  | Boolean |  False  | Whether the plugin should block attempts to install unlocked dependencies to the test environment. If enabled, then the [`tox_testenv_install_deps`](https://tox.readthedocs.io/en/latest/plugins.html#tox.hookspecs.tox_testenv_install_deps) plugin hook will be intercepted and an error will be raised if the test environment has the `deps` option configured. |\n| `install_project_deps` | Boolean |  True   | Whether all of the Poetry primary dependencies for the project package should be installed to the test environment.                                                                                                                                                                                                                                                  |\n| `require_poetry`       | Boolean |  False  | Whether Tox should be forced to fail if the plugin cannot import Poetry locally. If `False` then the plugin will be skipped for the test environment if Poetry cannot be imported. If `True` then the plugin will force the environment to error and the Tox run to fail.                                                                                            |\n| `poetry_dep_groups`    |  List   |  `[]`   | Names of Poetry dependency groups specified in `pyproject.toml` to install to the test environment.                                                                                                                                                                                                                                                                  |\n\n> **Note:** The `install_dev_deps` configuration option is deprecated and will be removed in\n> version 1.0.0. Please set `poetry_dep_groups = [dev]` in `tox.ini` for environments that\n> install the development dependencies.\n\n### Runtime Options\n\nAll arguments listed below can be passed to the `tox` command to modify runtime behavior\nof the plugin.\n\n| Argument                     |  Type   | Default | Description                                                                                                                                                                                                                                                                          |\n| :--------------------------- | :-----: | :-----: | :----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------- |\n| `--parallel-install-threads` | Integer |  `10`   | Number of worker threads to use to install dependencies in parallel. Installing in parallel with more threads can greatly speed up the install process, but can cause race conditions during install. Pass this option with the value `0` to entirely disable parallel installation. |\n\n> **Note:** The `--require-poetry` runtime option is deprecated and will be removed in\n> version 1.0.0. Please set `require_poetry = true` in `tox.ini` for environments that\n> should fail if Poetry is not available.\n\n> **Note:** The `--parallelize-locked-install` option is deprecated and will be removed in\n> version 1.0.0. Please use the `--parallel-install-threads` option.\n\n### Errors\n\nThere are several errors that the plugin can encounter for a test environment when Tox is\nrun. If an error is encountered then the status of the test environment that caused the\nerror will be set to one of the "Status" values below to indicate what the error was.\n\n| Status/Name                     | Cause                                                                                                                                                                                                                           |\n| :------------------------------ | :------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------ |\n| `ExtraNotFoundError`            | Indicates that the [`extras`](https://tox.readthedocs.io/en/latest/config.html#conf-extras) config option specified an extra that is not configured by Poetry in `pyproject.toml`.                                              |\n| `LockedDepVersionConflictError` | Indicates that an item in the `locked_deps` config option includes a [PEP-508 version specifier](https://www.python.org/dev/peps/pep-0508/#grammar) (ex: `pytest >=6.0, <6.1`).                                                 |\n| `LockedDepNotFoundError`        | Indicates that an item specified in the `locked_deps` config option does not match the name of a package in the Poetry lockfile.                                                                                                |\n| `LockedDepsRequiredError`       | Indicates that a test environment with the `require_locked_deps` config option set to `true` also specified unlocked dependencies using the [`deps`](https://tox.readthedocs.io/en/latest/config.html#conf-deps) config option. |\n| `PoetryNotInstalledError`       | Indicates that the `poetry` module could not be imported under the current runtime environment, and `require_poetry = true` was specified.                                                                                      |\n| `RequiresUnsafeDepError`        | Indicates that the package-under-test depends on a package that Poetry has classified as unsafe and cannot be installed.                                                                                                        |\n\n> **Note:** One or more of these errors can be caused by the `pyproject.toml` being out of\n> sync with the Poetry lockfile. If this is the case, than a warning will be logged when Tox\n> is run.\n\n### Other Notes\n\n#### Unsupported Tox config options\n\nBelow are the built-in Tox config options that are not respected by this plugin. All of\nthese options are made obsolete by the Poetry lockfile: either they aren\'t needed or their\nequivalent functionality is instead taken directly from the package details Poetry stores\nin its lockfile.\n\n> **Note:** The unsupported Tox config options will still apply to unlocked dependencies\n> being installed with the default Tox installation backend.\n\n- [`install_command`](https://tox.readthedocs.io/en/latest/config.html#conf-install_command)\n- [`pip_pre`](https://tox.readthedocs.io/en/latest/config.html#conf-pip_pre)\n- [`download`](https://tox.readthedocs.io/en/latest/config.html#conf-download)\n- [`indexserver`](https://tox.readthedocs.io/en/latest/config.html#conf-indexserver)\n- [`usedevelop`](https://tox.readthedocs.io/en/latest/config.html#conf-indexserver)\n\n#### Updating locked dependencies in a testenv\n\nWhen Poetry updates the version of a package in the lockfile (using either `poetry lock`\nor `poetry update`) then the plugin will automatically use this new version to install the\npackage to a test environment; there is no need to manually update `tox.ini` after\nupdating the Poetry lockfile.\n\nHowever, the plugin cannot determine when the lockfile is updated. If a Tox test\nenvironment has already been created then it will need to be recreated (using Tox\'s\nbuilt-in\n[`--recreate`](https://tox.readthedocs.io/en/latest/example/basic.html#forcing-re-creation-of-virtual-environments)\noption) for the new version to be found and installed.\n\n> **Note:** To force Tox to always recreate a test environment the\n> [`recreate`](https://tox.readthedocs.io/en/latest/config.html#conf-recreate) config option\n> can be set.\n\n#### Installing unsafe dependencies\n\nThere are several packages that cannot be installed from the lockfile because they are\nexcluded by Poetry itself. As a result these packages cannot be installed by this plugin\neither as test environment dependencies passed directly to `locked_deps` or as a transient\ndependency. When one of these packages is encountered by the plugin a warning will be\nlogged to the console and\n**the unsafe package will not be installed to the test environment**.\n\nThis list can be found in the Poetry source code\n[here](https://github.com/python-poetry/poetry/blob/master/poetry/puzzle/provider.py). As\nof [Poetry 1.1.6](https://github.com/python-poetry/poetry/releases/tag/1.1.6) there are\nfour packages classified as "unsafe" by Poetry and excluded from the lockfile:\n\n- `setuptools`\n- `distribute`\n- `pip`\n- `wheel`\n\n#### Using with an unmanaged Poetry installation\n\nIn CI/CD systems, automation environments, or other Python environments where the loaded\nsite packages are not managed by Poetry, it can be useful to manage the local installation\nof Poetry externally. This also helps to avoid problems that can be caused by the\n`--no-root`, `--no-dev`, or `--remove-untracked` arguments to the `poetry install` command\nwhich, in some situations, can cause Poetry to uninstall itself if Poetry is specified as\na dependency of one of the packages it is managing (like this plugin). To support these\nuse cases, this plugin specifies the `poetry` package as an optional dependency that can\nbe installed using a setuptools extra also named `poetry`.\n\n**Critical Warning: This plugin requires Poetry to function. If the plugin is installed without the `poetry` setuptools extra then Poetry must be installed independently for the plugin to function properly.**\n\nTo skip installing the `poetry` package as a dependency of `tox-poetry-installer`, do not\nspecify the `poetry` extra when adding the plugin:\n\n```bash\n# Adding the package without the "[poetry]" extra specifier so that\n# Poetry is not added as a transient dev-dependency:\npoetry add tox-poetry-installer --dev\n\n# Adding the package with the "[poetry]" extra specifier, so the Poetry\n# package will be added to the environment and tracked in the lockfile:\npoetry add tox-poetry-installer[poetry] --dev\n```\n\nOnce the plugin is installed- either with or without the Poetry extra- you can validate\nthat the plugin will run correctly with the following command. This command checks that\nall three required components (Tox, Poetry, and the plugin itself) are available in the\ncurrent Python environment:\n\n```bash\npython -c \'\\\n  import tox;\\\n  import tox_poetry_installer;\\\n  from poetry.poetry import Poetry;\\\n\'\n```\n\n> **Note:** To force Tox to fail if Poetry is not installed, add the `require_poetry = true`\n> option to the tox `[testenv]` configuration. See the\n> [Config Options](#configuration-options) for more information.\n\n## Developer Documentation\n\nAll project contributors and participants are expected to adhere to the\n[Contributor Covenant Code of Conduct, v2](CODE_OF_CONDUCT.md) ([external link](https://www.contributor-covenant.org/version/2/0/code_of_conduct/)).\n\nThe `devel` branch has the latest (and potentially unstable) changes. The stable releases\nare tracked on [Github](https://github.com/enpaul/tox-poetry-installer/releases),\n[PyPi](https://pypi.org/project/tox-poetry-installer/#history), and in the\n[Changelog](CHANGELOG.md).\n\n- To report a bug, request a feature, or ask for assistance, please\n  [open an issue on the Github repository](https://github.com/enpaul/tox-poetry-installer/issues/new).\n- To report a security concern or code of conduct violation, please contact the project\n  author directly at **\u200cme \\[at\u200c\\] enp dot\u200e \u200cone**.\n- To submit an update, please\n  [fork the repository](https://docs.github.com/en/enterprise/2.20/user/github/getting-started-with-github/fork-a-repo)\n  and [open a pull request](https://github.com/enpaul/tox-poetry-installer/compare).\n\nDeveloping this project requires [Python 3.7+](https://www.python.org/downloads/) and\n[Poetry 1.2](https://python-poetry.org/docs/#installation) or later. GNU Make can\noptionally be used to quickly setup a local development environment, but this is not\nrequired.\n\nTo setup a local development environment:\n\n```bash\n# Clone the repository...\n# ...over HTTPS\ngit clone https://github.com/enpaul/tox-poetry-installer.git\n# ...over SSH\ngit clone git@github.com:enpaul/tox-poetry-installer.git\n\ncd tox-poetry-installer/\n\n# Create and configure the local development environment\nmake dev\n\n# Run tests and CI locally\nmake test\n\n# See additional make targets\nmake help\n```\n\n> **Note:** The pre-commit hooks require dependencies in the Poetry environment to run. To\n> make a commit with the pre-commit hooks, you will need to run `poetry run git commit` or,\n> alternatively, [launch an environment shell](https://python-poetry.org/docs/cli/#shell).\n\n## Road Map\n\nThis project is under active development and is classified as beta software, ready for\nproduction environments on a provisional basis only.\n\n- Beta classification was assigned with\n  [v0.6.0](https://github.com/enpaul/tox-poetry-installer/releases/tag/0.6.0)\n- Stable classification will be assigned when the test suite covers an acceptable number of\n  use cases\n\n### Path to Beta\n\n- [x] Verify that primary package dependencies (from the `.package` env) are installed\n  correctly using the Poetry backend.\n- [x] Support the [`extras`](https://tox.readthedocs.io/en/latest/config.html#conf-extras) Tox\n  configuration option ([#4](https://github.com/enpaul/tox-poetry-installer/issues/4))\n- [x] Add per-environment Tox configuration option to fall back to default installation\n  backend.\n- [ ] ~Add warnings when an unsupported Tox configuration option is detected while using the\n  Poetry backend.~ ([#5](https://github.com/enpaul/tox-poetry-installer/issues/5))\n- [x] Add trivial tests to ensure the project metadata is consistent between the pyproject.toml\n  and the module constants.\n- [x] Update to use [poetry-core](https://github.com/python-poetry/poetry-core) and improve\n  robustness of the Tox and Poetry module imports to avoid potentially breaking API changes\n  in upstream packages. ([#2](https://github.com/enpaul/tox-poetry-installer/issues/2))\n- [ ] ~Find and implement a way to mitigate the\n  [UNSAFE_DEPENDENCIES issue](https://github.com/python-poetry/poetry/issues/1584) in\n  Poetry.~ ([#6](https://github.com/enpaul/tox-poetry-installer/issues/6))\n- [x] Fix logging to make proper use of Tox\'s logging reporter infrastructure ([#3](https://github.com/enpaul/tox-poetry-installer/issues/3))\n- [x] Add configuration option for installing all dev-dependencies to a testenv ([#14](https://github.com/enpaul/tox-poetry-installer/issues/14))\n\n### Path to Stable\n\nEverything in Beta plus...\n\n- [ ] Fully replace dependency on `poetry` with dependency on `poetry-core` ([#2](https://github.com/enpaul/tox-poetry-installer/issues/2))\n- [x] Add comprehensive unit tests\n- [ ] ~Add tests for each feature version of Tox between 3.8 and 3.20~\n- [x] Add tests for Python-3.6, 3.7, 3.8, and 3.9\n- [x] Add Github Actions based CI\n',
    'author': 'Ethan Paul',
    'author_email': '24588726+enpaul@users.noreply.github.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/enpaul/tox-poetry-installer/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""Static constants for reference

Rule of thumb: if it's an arbitrary value that will never be changed at runtime, it should go
in this module.

All constants should be type hinted.
"""
from typing import Set
from typing import Tuple

from tox_poetry_installer import __about__


# Valid PEP508 version delimiters. These are used to test whether a given string (specifically a
# dependency name) is just a package name or also includes a version identifier.
PEP508_VERSION_DELIMITERS: Tuple[str, ...] = ("~=", "==", "!=", ">", "<")

# Prefix all reporter messages should include to indicate that they came from this module in the
# console output.
REPORTER_PREFIX: str = f"{__about__.__title__}:"

# Internal list of packages that poetry has deemed unsafe and are excluded from the lockfile
# TODO: This functionality is no longer needed, should remove in a future update.
UNSAFE_PACKAGES: Set[str] = set()

# Number of threads to use for installing dependencies by default
DEFAULT_INSTALL_THREADS: int = 10



"""Custom plugin exceptions

All exceptions should inherit from the common base exception :exc:`ToxPoetryInstallerException`.

::

  ToxPoetryInstallerException
   +-- SkipEnvironment
   |    +-- PoetryNotInstalledError
   +-- LockedDepVersionConflictError
   +-- LockedDepNotFoundError
   +-- ExtraNotFoundError
   +-- LockedDepsRequiredError
   +-- RequiresUnsafeDepError

"""


class ToxPoetryInstallerException(Exception):
    """Error while installing locked dependencies to the test environment"""


class SkipEnvironment(ToxPoetryInstallerException):
    """Current environment does not meet preconditions and should be skipped by the plugin"""


class PoetryNotInstalledError(SkipEnvironment):
    """No version of Poetry could be imported from the current Python environment"""


class LockedDepVersionConflictError(ToxPoetryInstallerException):
    """Locked dependencies cannot specify an alternate version for installation"""


class LockedDepNotFoundError(ToxPoetryInstallerException):
    """Locked dependency was not found in the lockfile"""


class ExtraNotFoundError(ToxPoetryInstallerException):
    """Project package extra not defined in project's pyproject.toml"""


class LockedDepsRequiredError(ToxPoetryInstallerException):
    """Environment cannot specify unlocked dependencies when locked dependencies are required"""


class RequiresUnsafeDepError(ToxPoetryInstallerException):
    """Package under test depends on an unsafe dependency and cannot be installed"""
